@@ -296,6 +296,9 @@ def test_WorkerTable(c, s, a, b):
     assert all(wt.source.data.values())
     assert all(len(v) == 2 for v in wt.source.data.values())
 
+    ncores = wt.source.data['ncores']
+    assert all(ncores)
+
 
 @gen_cluster(client=True)
 def test_WorkerTable_custom_metrics(c, s, a, b):
