@@ -712,7 +712,10 @@ def test_inproc_comm_closed_explicit_2():
     comm = yield connect(contact_addr)
     comm.close()
     assert comm.closed()
-    yield gen.sleep(0.01)
+    start = time()
+    while len(listener_errors) < 1:
+        assert time() < start + 1
+        yield gen.sleep(0.01)
     assert len(listener_errors) == 1
 
     with pytest.raises(CommClosedError):
@@ -730,8 +733,11 @@ def test_inproc_comm_closed_explicit_2():
 
     comm = yield connect(contact_addr)
     comm.write("foo")
-    yield gen.sleep(0.01)
-    assert comm.closed()
+
+    start = time()
+    while not comm.closed():
+        yield gen.sleep(0.01)
+        assert time() < start + 2
 
     comm.close()
     comm.close()
@@ -929,6 +935,7 @@ def check_deserialize(addr):
     yield check_connector_deserialize(addr, True, msg, partial(check_out, True))
 
 
+@pytest.mark.xfail(reason='intermittent failure on windows')
 @gen_test()
 def test_tcp_deserialize():
     yield check_deserialize('tcp://')
